import re
from typing import Any, Dict, Union, TextIO, Hashable

import yaml as yaml_library
import importlib.resources as pkg_resources
from . import templates


_path_matcher = re.compile(r'\$\{([^}^{]+)\}')

DataType = Dict[str, Any]
YamlStrType = Union[str, TextIO]
YamlType = Union[Dict[Hashable, Any], list, None]


def _path_constructor_factory(data: DataType, strict: bool):
    """Constructor factory for yaml.

    A factory that creates a constructor method, that replaces yaml-variables with the values in the data dictionary.

    :param data: dictionary that replaces yaml-variables.
    :param strict: if this is False all yaml variables that are not in the data dictionary will have a default value.
    :return: yaml constructor.
    """
    def path_constructor(loader, node):
        value = node.value
        match = _path_matcher.match(value)
        var = match.group()[2:-1]
        if strict:
            # raise exception if key is not found
            val = data[var]
        else:
            # default value if key is not found
            val = data.get(var)
        # needed to prevent converting integers to strings
        if value[match.end():] == "":
            return val
        else:
            return str(val) + value[match.end():]
    return path_constructor


class _VariableLoader(yaml_library.FullLoader):
    """Loader that is used for replacing yaml-variables."""
    yaml_constructors = yaml_library.FullLoader.yaml_constructors.copy()
    yaml_implicit_resolvers = yaml_library.FullLoader.yaml_implicit_resolvers.copy()


class TemplateEngine:
    """Yaml Template Engine.

    Used to replace yaml-variables.
    """

    def __init__(self, yaml_lib=yaml_library):
        self.yaml_lib = yaml_lib

    def load(self, yaml_str: YamlStrType, data: DataType, strict: bool = False) -> YamlType:
        """Parses a yaml string to a python object and replaces yaml-variables.

        :param yaml_str: The yaml string that should be parsed.
        :param data: The data that should be inserted into the yaml-variables.
        :param strict: If True, an error will be thrown when variables have no value in the data dictionary. If false
            the default value None will be used.
        :return: yaml object.
        """
        _VariableLoader.add_implicit_resolver('!path', _path_matcher, None)
        _VariableLoader.add_constructor('!path', _path_constructor_factory(data, strict))
        p = self.yaml_lib.load(yaml_str, Loader=_VariableLoader)
        return p

    def load_template(self, template: str, data: DataType, strict: bool = False) -> YamlType:
        """Reads a template from the template module and parses the content as yaml to a python object and replaces
        yaml-variables.

        :param template: The template that contains the yaml.
        :param data: The data that should be inserted into the yaml-variables.
        :param strict: If True, an error will be thrown when variables have no value in the data dictionary. If false
        the default value None will be used.
        :return: yaml object.
        """
        cont = pkg_resources.read_text(templates, template)
        return self.load(cont, data, strict)

    def load_file(self, filename: str, data: DataType, strict: bool = False) -> YamlType:
        """Reads a file and parses the content as yaml to a python object and replaces yaml-variables.

        :param filename: The file that contains the yaml.
        :param data: The data that should be inserted into the yaml-variables.
        :param strict: If True, an error will be thrown when variables have no value in the data dictionary. If false
        the default value None will be used.
        :return: yaml object.
        """
        with open(filename) as cont:
            return self.load(cont, data, strict)

    def dump(self, yaml: Union[YamlType, Any]) -> str:
        """Converts a yaml object back to a string."""
        return self.yaml_lib.dump(yaml, Dumper=self.yaml_lib.Dumper, allow_unicode=True)

    def replace_template(self, template: str, data: DataType, strict: bool = False) -> str:
        """Reads a template and replaces the variables.

        :param template: The template that contains the yaml.
        :param data: The data that should be inserted into the yaml-variables.
        :param strict: If True, an error will be thrown when variables have no value in the data dictionary. If false
        the default value None will be used.
        :return: The file content as string with replaced variables.
        """
        yaml = self.load_template(template, data, strict)
        return self.dump(yaml)

    def replace_file(self, filename: str, data: DataType, strict: bool = False) -> str:
        """Reads a file and replaces the variables.

        :param filename: The file that contains the yaml.
        :param data: The data that should be inserted into the yaml-variables.
        :param strict: If True, an error will be thrown when variables have no value in the data dictionary. If false
        the default value None will be used.
        :return: The file content as string with replaced variables.
        """
        yaml = self.load_file(filename, data, strict)
        return self.dump(yaml)

    def replace(self, yaml_str: YamlStrType, data: DataType) -> str:
        """Replaces the variables in the yaml string.

        :param yaml_str: The yaml string that should be replaced.
        :param data: The data that should be inserted into the yaml-variables.
        :param strict: If True, an error will be thrown when variables have no value in the data dictionary. If false
        the default value None will be used.
        :return: The yaml string with replaced variables.
        """
        yaml = self.load(yaml_str, data)
        return self.dump(yaml)



from lab_orchestrator_lib.kubernetes.api import APIRegistry, NamespacedApi, NotNamespacedApi
from lab_orchestrator_lib.template_engine import TemplateEngine


class KubernetesController:
    template_file = None

    def __init__(self, registry: APIRegistry):
        self.registry = registry

    def _get_template(self, template_data) -> str:
        return TemplateEngine().replace_template(template=self.template_file, data=template_data)


class NamespacedController(KubernetesController):
    def _api(self) -> NamespacedApi:
        raise NotImplementedError()

    def get_list(self, namespace) -> str:
        return self._api().get_list(namespace)

    def get(self, namespace, identifier) -> str:
        return self._api().get(namespace, identifier)

    def delete(self, namespace, identifier) -> str:
        return self._api().delete(namespace, identifier)


class NotNamespacedController(KubernetesController):
    def _api(self) -> NotNamespacedApi:
        raise NotImplementedError()

    def get_list(self):
        return self._api().get_list()

    def get(self, identifier):
        return self._api().get(identifier)

    def delete(self, identifier):
        return self._api().delete(identifier)


from typing import Generic, List, TypeVar, Any

from lab_orchestrator_lib.model.model import DockerImage, Lab, LabInstance

Adapter = TypeVar('Adapter')
LibModelType = TypeVar('LibModelType', DockerImage, Lab, LabInstance)  # subclasses of Model


class AdapterController(Generic[Adapter, LibModelType]):
    def __init__(self, adapter: Adapter):
        self.adapter = adapter

    def get_all(self) -> List[LibModelType]:
        return self.adapter.get_all()

    def get(self, identifier) -> LibModelType:
        return self.adapter.get(identifier)

    def get_by_attr(self, attr, value) -> LibModelType:
        return self.adapter.get_by_attr(attr, value)

    def delete(self, identifier) -> None:
        return self.adapter.delete(identifier)

    def save(self, obj: LibModelType) -> LibModelType:
        return self.adapter.save(obj)

    def filter(self, **kwargs) -> LibModelType:
        return self.adapter.filter(**kwargs)


from typing import Union

from lab_orchestrator_lib.custom_exceptions import ValidationError


Identifier = Union[str, int]


class Model:
    def __init__(self, primary_key: Identifier):
        self.primary_key = primary_key


class User(Model):
    def __init__(self, primary_key: Identifier):
        super().__init__(primary_key)


class DockerImage(Model):
    def __init__(self, primary_key: Identifier, name: str, description: str, url: str):
        super().__init__(primary_key)
        if len(name) > 32:
            raise ValidationError("name is longer than 32 characters.")
        if len(description) > 128:
            raise ValidationError("description is longer than 128 characters.")
        if len(url) > 256:
            raise ValidationError("url is longer than 256 characters.")
        self.name = name
        self.description = description
        self.url = url


class Lab(Model):
    def __init__(self, primary_key: Identifier, name: str, namespace_prefix: str, description: str,
                 docker_image_id: Identifier, docker_image_name: str):
        super().__init__(primary_key)
        if len(name) > 32:
            raise ValidationError("name is longer than 32 characters.")
        if len(namespace_prefix) > 32:
            raise ValidationError("namespace_prefix is longer than 32 characters.")
        if len(description) > 128:
            raise ValidationError("description is longer than 128 characters.")
        if len(docker_image_name) > 32:
            raise ValidationError("docker_image_name is longer than 32 characters.")
        self.name = name
        self.namespace_prefix = namespace_prefix
        self.description = description
        self.docker_image_id = docker_image_id
        self.docker_image_name = docker_image_name


class LabInstance(Model):
    def __init__(self, primary_key: Identifier, lab_id: Identifier, user_id: Identifier):
        super().__init__(primary_key)
        self.lab_id = lab_id
        self.user_id = user_id


class LabInstanceKubernetes(Model):
    """Doesn't need any adapter and should not be saved in the database. This is used to return information about a
    lab instance including information from kubernetes and the jwt token."""
    def __init__(self, primary_key: Identifier, lab_id: Identifier, user_id: Identifier, jwt_token: str):
        super().__init__(primary_key)
        self.lab_id = lab_id
        self.user_id = user_id
        self.jwt_token = jwt_token





from typing import List

from lab_orchestrator_lib_auth.auth import generate_auth_token, LabInstanceTokenParams
from lab_orchestrator_lib.controller.adapter_controller import AdapterController
from lab_orchestrator_lib.controller.kubernetes_controller import NamespacedController, NotNamespacedController
from lab_orchestrator_lib.database.adapter import DockerImageAdapterInterface, LabAdapterInterface, \
    LabInstanceAdapterInterface, UserAdapterInterface
from lab_orchestrator_lib.kubernetes.api import NotNamespacedApi, NamespacedApi, APIRegistry
from lab_orchestrator_lib.model.model import DockerImage, Lab, LabInstance, Identifier, User, LabInstanceKubernetes


class UserController:
    def __init__(self, adapter: UserAdapterInterface):
        self.adapter = adapter

    def get_all(self) -> List[User]:
        return self.adapter.get_all()

    def get(self, identifier) -> User:
        return self.adapter.get(identifier)


class NamespaceController(NotNamespacedController):
    template_file = "namespace_template.yaml"

    def _api(self) -> NotNamespacedApi:
        return self.registry.namespace

    def create(self, namespace):
        template_data = {'namespace': namespace}
        data = self._get_template(template_data)
        return self._api().create(data)


class NetworkPolicyController(NamespacedController):
    template_file = 'network_policy_template.yaml'

    def _api(self) -> NamespacedApi:
        return self.registry.network_policy

    def __init__(self, registry: APIRegistry):
        super().__init__(registry)
        self.default_name = "allow-same-namespace"

    def create(self, namespace):
        template_data = {'namespace': namespace, 'network_policy_name': self.default_name}
        data = self._get_template(template_data)
        return self._api().create(namespace, data)


class DockerImageController(AdapterController):
    def __init__(self, adapter: DockerImageAdapterInterface):
        super().__init__(adapter)

    def create(self, name, description, url) -> DockerImage:
        return self.adapter.create(name, description, url)


class VirtualMachineInstanceController(NamespacedController):
    template_file = 'vmi_template.yaml'

    def __init__(self, registry: APIRegistry, namespace_ctrl: NamespaceController,
                 docker_image_ctrl: DockerImageController):
        super().__init__(registry)
        self.namespace_ctrl = namespace_ctrl
        self.docker_image_ctrl = docker_image_ctrl

    def _api(self) -> NamespacedApi:
        return self.registry.virtual_machine_instance

    def create(self, namespace, lab: Lab):
        docker_image = self.docker_image_ctrl.get(lab.docker_image_id)
        template_data = {"cores": 3, "memory": "3G",
                         "vm_image": docker_image.url, "vmi_name": lab.docker_image_name,
                         "namespace": namespace}
        data = self._get_template(template_data)
        return self._api().create(namespace, data)

    def get_list_of_lab_instance(self, lab_instance: LabInstance, lab_instance_ctrl: 'LabInstanceController'):
        namespace_name = LabInstanceController.get_namespace_name(lab_instance, lab_instance_ctrl)
        namespace = self.namespace_ctrl.get(namespace_name)
        return self.get_list(namespace_name)

    def get_of_lab_instance(self, lab_instance: LabInstance, virtual_machine_instance_id,
                            lab_instance_ctrl: 'LabInstanceController'):
        namespace_name = LabInstanceController.get_namespace_name(lab_instance, lab_instance_ctrl)
        namespace = self.namespace_ctrl.get(namespace_name)
        return self.get(namespace_name, virtual_machine_instance_id)


class LabController(AdapterController):
    def __init__(self, adapter: LabAdapterInterface):
        super().__init__(adapter)

    def create(self, name: str, namespace_prefix: str, description: str, docker_image_id: Identifier,
               docker_image_name: str) -> Lab:
        return self.adapter.create(name=name, namespace_prefix=namespace_prefix, description=description,
                                   docker_image_id=docker_image_id, docker_image_name=docker_image_name)


class LabInstanceController(AdapterController):
    def __init__(self,
                 adapter: LabInstanceAdapterInterface,
                 virtual_machine_instance_ctrl: VirtualMachineInstanceController,
                 namespace_ctrl: NamespaceController,
                 lab_ctrl: LabController,
                 network_policy_ctrl: NetworkPolicyController,
                 user_ctrl: UserController,
                 secret_key: str):
        super().__init__(adapter)
        self.virtual_machine_instance_ctrl = virtual_machine_instance_ctrl
        self.namespace_ctrl = namespace_ctrl
        self.lab_ctrl = lab_ctrl
        self.network_policy_ctrl = network_policy_ctrl
        self.user_ctrl = user_ctrl
        self.secret_key = secret_key

    @staticmethod
    def get_namespace_name(lab_instance: LabInstance, lab_instance_ctrl):
        lab = lab_instance_ctrl.get(lab_instance.lab_id)
        return LabInstanceController.gen_namespace_name(lab, lab_instance.user_id, lab_instance.primary_key)

    @staticmethod
    def gen_namespace_name(lab: Lab, user_id, lab_instance_id):
        return f"{lab.namespace_prefix}-{user_id}-{lab_instance_id}"

    def create(self, lab_id: Identifier, user_id: Identifier) -> LabInstanceKubernetes:
        lab = self.lab_ctrl.get(lab_id)
        if lab is None:
            # TODO sinnvolle exception werfen
            raise Exception
        user = self.user_ctrl.get(user_id)
        if user is None:
            # TODO sinnvolle exception werfen
            raise Exception
        lab_instance = self.adapter.create(lab_id=lab_id, user_id=user_id)
        # create namespace
        namespace_name = LabInstanceController.gen_namespace_name(lab, user_id, lab_instance.primary_key)
        namespace = self.namespace_ctrl.create(namespace_name)
        # TODO fix response code
        # TODO log if deletion doesn't work
        #if namespace.response_code != 0:
        #    self.adapter.delete(lab_instance.primary_key)
        #    raise Exception
        # create network policy
        network_policy = self.network_policy_ctrl.create(namespace_name)
        #if network_policy.response_code != 0:
        #    self.adapter.delete(lab_instance.primary_key)
        #    self.namespace_ctrl.delete(namespace_name)
        #    raise Exception
        # create vmi
        vmi = self.virtual_machine_instance_ctrl.create(namespace_name, lab)
        #if vmi.response_code != 0:
        #    self.adapter.delete(lab_instance.primary_key)
        #    self.namespace_ctrl.delete(namespace_name)
        #    raise Exception
        lab_instance_token_params = LabInstanceTokenParams(lab_id, namespace_name, lab.docker_image_name)
        token = generate_auth_token(user_id=user_id, lab_instance_token_params=lab_instance_token_params,
                                    secret_key=self.secret_key)
        return LabInstanceKubernetes(primary_key=lab_instance.primary_key, lab_id=lab_id, user_id=user_id, jwt_token=token)

    def delete(self, lab_instance: LabInstance):
        lab = self.lab_ctrl.get(lab_instance.lab_id)
        namespace_name = LabInstanceController.gen_namespace_name(lab, lab_instance.user_id, lab_instance.primary_key)
        # this also deletes VMIs and all other resources in the namespace
        self.namespace_ctrl.delete(namespace_name)
        # now delete local object
        super().delete(lab_instance.primary_key)

    def get_list_of_user(self, user: User):
        # TODO list instead of item
        lab_instances = self.adapter.filter_by(user_id=user.primary_key)
        return lab_instances

    def save(self, obj: LabInstance) -> LabInstance:
        raise Exception("LabInstances can't be mutated.")



import logging
from abc import ABC
from typing import Dict, Type, Callable, Union, Optional, Any

import requests

_API_EXTENSIONS_NAMESPACED: Dict[str, Type['NamespacedApi']] = {}
_API_EXTENSIONS_NOT_NAMESPACED: Dict[str, Type['NotNamespacedApi']] = {}


NamespacedApiDecorator = Callable[[Type['NamespacedApi'], ], Type['NamespacedApi']]
NotNamespacedApiDecorator = Callable[[Type['NotNamespacedApi'], ], Type['NotNamespacedApi']]


def add_api_namespaced(name: str) -> NamespacedApiDecorator:
    def inner(cls: Type[NamespacedApi]) -> Type[NamespacedApi]:
        _API_EXTENSIONS_NAMESPACED[name] = cls
        return cls
    return inner


def add_api_not_namespaced(name: str) -> NotNamespacedApiDecorator:
    def inner(cls: Type[NotNamespacedApi]) -> Type[NotNamespacedApi]:
        _API_EXTENSIONS_NOT_NAMESPACED[name] = cls
        return cls
    return inner


class Proxy:
    def __init__(self, base_uri: str, service_account_token: str = None,
                 cacert: str = None, insecure_ssl: str = False, requests_lib=requests):
        self.requests = requests_lib
        if service_account_token is None:
            logging.warning("No service account token.")
        if cacert is None:
            logging.warning("No cacert.")
        self.base_uri = base_uri
        self.service_account_token = service_account_token
        if insecure_ssl:
            self.verify = False
        elif cacert is None:
            self.verify = True
        else:
            self.verify = cacert

    def get(self, address: str) -> str:
        headers = {"Authorization": f"Bearer {self.service_account_token}"}
        response = self.requests.get(self.base_uri + address,
                                     headers=headers, verify=self.verify)
        return response.text

    def post(self, address: str, data: str) -> str:
        headers = {"Authorization": f"Bearer {self.service_account_token}",
                   "Content-Type": "application/yaml"}
        response = self.requests.post(self.base_uri + address,
                                      data=data, headers=headers, verify=self.verify)
        return response.text

    def delete(self, address) -> str:
        headers = {"Authorization": f"Bearer {self.service_account_token}"}
        response = self.requests.delete(self.base_uri + address,
                                        headers=headers, verify=self.verify)
        return response.text


class APIRegistry:
    def __init__(self, proxy: Proxy):
        self.proxy = proxy

    def __dir__(self):
        keys = set(super().__dir__())
        keys = keys.union(_API_EXTENSIONS_NAMESPACED.keys())
        keys = keys.union(_API_EXTENSIONS_NOT_NAMESPACED.keys())
        return list(keys)

    def __getattr__(self, name) -> Union['NamespacedApi', 'NotNamespacedApi']:
        cls: Union[Optional[Type['NamespacedApi']], Optional[Type['NotNamespacedApi']]]
        if cls := _API_EXTENSIONS_NAMESPACED.get(name):
            return cls(self.proxy)
        if cls := _API_EXTENSIONS_NOT_NAMESPACED.get(name):
            return cls(self.proxy)
        raise AttributeError(f'{name} not found')


class ApiExtension(ABC):
    """Used to extend the APIRegistry.

    To extend the APIRegistry, you need to implement either the NamespacedApi or NotNamespacedApi. Then add one
    of the decorators add_api_namespaced or add_api_not_namespaced or use these methods to add the classes manually.
    After that you can access an instance of the extension through the proxy as attributes for example
    "proxy.namespace" or "proxy.virtual_machine_instance". With that instances you are able to access the Kubernetes
    Api with the methods in NamespacedApi and NotNamespacedApi for example get_list or create.

    :param list_url: Will be formatted. Which variables will be inserted depends on the ApiExtension type.
    :param details_url: Will be formatted. Which variables will be inserted depends on the ApiExtension type.
    """
    list_url = None
    detail_url = None

    def __init__(self, proxy: Proxy):
        self.proxy = proxy


class NamespacedApi(ApiExtension):
    """Abstract base class extension for resources that are namespaced.

    :list_url: Will be formated with the variable "namespace".
    :detail_url: Will be formated with the variable "namespace" and "identifier".
    """
    def get_list(self, namespace: str) -> str:
        return self.proxy.get(self.list_url.format(namespace=namespace))

    def create(self, namespace: str, data: str) -> str:
        return self.proxy.post(self.list_url.format(namespace=namespace), data)

    def get(self, namespace: str, identifier: str) -> str:
        return self.proxy.get(self.detail_url.format(namespace=namespace, identifier=identifier))

    def delete(self, namespace: str, identifier: str) -> str:
        return self.proxy.delete(self.detail_url.format(namespace=namespace, identifier=identifier))


class NotNamespacedApi(ApiExtension):
    """Abstract base class extension for resources that are not namespaced.

    :list_url: Will not be formated and shouldn't contain variables.
    :detail_url: Will be formated with the variable "identifier".
    """

    def get_list(self) -> str:
        return self.proxy.get(self.list_url)

    def create(self, data: str) -> str:
        return self.proxy.post(self.list_url, data)

    def get(self, identifier: str) -> str:
        return self.proxy.get(self.detail_url.format(identifier=identifier))

    def delete(self, identifier: str) -> str:
        return self.proxy.delete(self.detail_url.format(identifier=identifier))


@add_api_not_namespaced("namespace")
class Namespace(NotNamespacedApi):
    list_url = "/api/v1/namespaces"
    detail_url = "/api/v1/namespaces/{identifier}"


@add_api_namespaced("virtual_machine_instance")
class VirtualMachineInstance(NamespacedApi):
    list_url = "/apis/kubevirt.io/v1alpha3/namespaces/{namespace}/virtualmachineinstances/"
    detail_url = "/apis/kubevirt.io/v1alpha3/namespaces/{namespace}/virtualmachineinstances/{identifier}"


@add_api_namespaced("network_policy")
class NetworkPolicy(NamespacedApi):
    list_url = "/apis/networking.k8s.io/v1/namespaces/{namespace}/networkpolicies"
    detail_url = "/apis/networking.k8s.io/v1/namespaces/{namespace}/networkpolicies/{identifier}"



import os
from dataclasses import dataclass
from typing import Optional

from lab_orchestrator_lib.kubernetes.api import Proxy, APIRegistry


@dataclass
class KubernetesConfig:
    service_account_token: Optional[str]
    cacert: Optional[str]
    protocol: str
    service_host: str
    service_port: str
    base_uri: str


def get_kubernetes_config():
    with open('/var/run/secrets/kubernetes.io/serviceaccount/token', 'r') as file:
        service_account_token = file.read()
    cacert = '/var/run/secrets/kubernetes.io/serviceaccount/ca.crt'
    protocol = "https"
    kubernetes_service_host = os.environ["KUBERNETES_SERVICE_HOST"]
    kubernetes_service_port = os.environ["KUBERNETES_SERVICE_PORT"]
    base_uri = f"{protocol}://{kubernetes_service_host}:{kubernetes_service_port}"
    kubernetes_config = KubernetesConfig(service_account_token, cacert, protocol, kubernetes_service_host,
                                         kubernetes_service_port, base_uri)
    return kubernetes_config


def get_development_config():
    service_account_token = None
    cacert = None
    protocol = "http"
    kubernetes_service_host = os.environ["KUBERNETES_SERVICE_HOST"]
    kubernetes_service_port = os.environ["KUBERNETES_SERVICE_PORT"]
    base_uri = f"{protocol}://{kubernetes_service_host}:{kubernetes_service_port}"
    kubernetes_config = KubernetesConfig(service_account_token, cacert, protocol, kubernetes_service_host,
                                         kubernetes_service_port, base_uri)
    return kubernetes_config


def get_registry(kubernetes_config: KubernetesConfig):
    proxy = Proxy(kubernetes_config.base_uri, kubernetes_config.service_account_token, kubernetes_config.cacert)
    return APIRegistry(proxy)



from typing import List, Any, Dict

from lab_orchestrator_lib.model.model import DockerImage, Lab, LabInstance, Identifier, User


class UserAdapterInterface:
    """

    Contains only get methods, because the lab_controller_lib won't create or delete users.
    """
    def get_all(self) -> List[User]:
        raise NotImplementedError()

    def get(self, identifier: Identifier) -> User:
        raise NotImplementedError()


class DockerImageAdapterInterface:
    def create(self, name: str, description: str, url: str) -> DockerImage:
        raise NotImplementedError()

    def get_all(self) -> List[DockerImage]:
        raise NotImplementedError()

    def get(self, identifier: Identifier) -> DockerImage:
        raise NotImplementedError()

    def get_by_attr(self, attr: str, value: Any) -> DockerImage:
        raise NotImplementedError()

    def delete(self, identifier: Identifier) -> None:
        raise NotImplementedError()

    def save(self, obj: DockerImage) -> DockerImage:
        raise NotImplementedError()

    def filter(self, **kwargs: Dict[str, Any]) -> DockerImage:
        raise NotImplementedError()


class LabAdapterInterface:
    def create(self, name: str, namespace_prefix: str, description: str, docker_image_id: Identifier,
               docker_image_name: str) -> Lab:
        raise NotImplementedError()

    def get_all(self) -> List[Lab]:
        raise NotImplementedError()

    def get(self, identifier: Identifier) -> Lab:
        raise NotImplementedError()

    def get_by_attr(self, attr: str, value: Any) -> Lab:
        raise NotImplementedError()

    def delete(self, identifier: Identifier) -> None:
        raise NotImplementedError()

    def save(self, obj: Lab) -> Lab:
        raise NotImplementedError()

    def filter(self, **kwargs: Dict[str, Any]) -> Lab:
        raise NotImplementedError()


class LabInstanceAdapterInterface:
    def create(self, lab_id: Identifier, user_id: Identifier) -> LabInstance:
        raise NotImplementedError()

    def get_all(self) -> List[LabInstance]:
        raise NotImplementedError()

    def get(self, identifier: Identifier) -> LabInstance:
        raise NotImplementedError()

    def get_by_attr(self, attr: str, value: Any) -> LabInstance:
        raise NotImplementedError()

    def delete(self, identifier: Identifier) -> None:
        raise NotImplementedError()

    def save(self, obj: LabInstance) -> LabInstance:
        raise NotImplementedError()

    def filter(self, **kwargs: Dict[str, Any]) -> LabInstance:
        raise NotImplementedError()
